"""
Module containing code for creating the command line argument parser.
"""
import argparse
from itertools import chain
import logging

from gisht import __version__
from gisht.args.autocomplete import gist_completer
from gisht.args.data import GistCommand


__all__ = ['create_argv_parser']


def create_argv_parser():
    """Create a :class:`argparse.ArgumentParser` object
    for parsing command line arguments passed by the user.
    """
    parser = argparse.ArgumentParser(
        description="Download & run GitHub gists straight from the terminal",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False)

    add_gist_group(parser)
    add_gist_command_group(parser)
    add_logging_group(parser)

    misc_group = add_misc_group(parser)

    # get the autogenerated usage string and tweak it a little
    # to include gist arguments that are handled separately
    # and exclude the miscellaneous flags which aren't part of a normal usage
    usage = parser.format_usage()
    usage = usage[usage.find(parser.prog):].rstrip("\n")  # remove cruft
    for misc_flag in chain.from_iterable(a.option_strings
                                         for a in misc_group._group_actions):
        usage = usage.replace(" [%s]" % misc_flag, "")
    parser.usage = usage + " [-- GIST_ARGS]"
    return parser


# Gist specification

def add_gist_group(parser):
    """Include an argument group that allows to specify the gist.
    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    group = parser.add_argument_group(
        "Gist", "Specifies the gist, optionally with flags")

    group.add_argument('gist', type=gist,
                       help="GitHub gist, specified as <owner>/<name> "
                            "(e.g. Octocat/foo)",
                       metavar="GIST").completer = gist_completer

    fetch_group = group.add_mutually_exclusive_group()
    fetch_group.set_defaults(local=None)
    fetch_group.add_argument('-l', '--local', '--cached',
                             action='store_true', dest='local',
                             help="operate only on gists available locally"
                                  " (do not fetch anything from GitHub)")
    fetch_group.add_argument('-f', '--fetch', '--remote',
                             action='store_false', dest='local',
                             help="always fetch the gist from GitHub, "
                                  "possibly updating it to latest version")

    return group


def gist(value):
    """Converter/validator for the GIST command line argument."""
    try:
        owner, gist_name = value.split('/')
    except ValueError:
        raise argparse.ArgumentTypeError(
            "%r is not a valid gist reference; try '<owner>/`<name>`" % value)
    if owner and gist_name:
        return value
    else:
        raise argparse.ArgumentTypeError(
            "neither gist owner or name can be empty (got %r)" % (value,))


# Gist command

def add_gist_command_group(parser):
    """Include an argument group that allows to specify
    the command to be performed on the gist.

    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    # 'command' is our internal wording, intended not to clash with
    # ArgumentParser actions, but the UI shall refer to it still as 'action'
    group = parser.add_argument_group(
        "Actions", "Possible actions to perform on the gist") \
        .add_mutually_exclusive_group()
    group.set_defaults(command=GistCommand.RUN)

    gist_commands = {
        GistCommand.RUN: "run specified gist; this is the default behavior "
                         "if no action was specified explicitly",
        GistCommand.WHICH: "output the path to binary which would be "
                           "ran for given gist; useful for passing it "
                           "to other commands via $( )",
        GistCommand.PRINT: "print gist source on the standard output",
        GistCommand.OPEN: "open the gist's GitHub page "
                          "in the default web browser",
        GistCommand.INFO: "show summary information about specified gist",
    }
    for cmd, help in gist_commands.items():
        group.add_argument(
            *cmd.flags, dest='command', action=GistCommandAction, help=help)

    return group


class GistCommandAction(argparse.Action):
    """Custom argument parser's :class:`Action` for handling
    gist command flags.
    """
    def __init__(self, **kwargs):
        option_strings = kwargs.get('option_strings')
        if not option_strings:
            raise TypeError("GistCommandAction can only be applied to flags")

        # basing upon supplied flag names, find the matching GistCommannd
        commands = set(map(GistCommand.for_flag, option_strings))
        if not commands:
            raise ValueError(
                "no gist command found matching flags %r" % (option_strings,))
        if len(commands) > 1:
            raise ValueError(
                "ambiguous gist command for flags %r -- can be one of: %r" % (
                    option_strings, commands))
        kwargs['const'] = commands.pop()

        super(GistCommandAction, self).__init__(nargs=0, **kwargs)

        #: Whether we have already seen a gist command flag.
        #: Used to prevent the same flag from being supplied twice.
        self._seen = False

    def __call__(self, parser, namespace, values, option_string=None):
        if self._seen:
            raise argparse.ArgumentError(self, "can be specified at most once")
        setattr(namespace, self.dest, self.const)
        self._seen = True


# Logging options

def add_logging_group(parser):
    """Include an argument group that allows to
    control verbosity of the program's logging output.

    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    group = parser.add_argument_group(
        "Verbosity", "Only errors and warnings are printed by default") \
        .add_mutually_exclusive_group()
    group.set_defaults(log_level=LogLevelAction.DEFAULT_LEVEL)

    group.add_argument('-v', '--verbose', dest='log_level',
                       action=LogLevelAction,
                       const=-LogLevelAction.DEFAULT_INCREMENT,
                       help="include finer grained details in the output; "
                            "this option can be specified multiple times")
    group.add_argument('-q', '--quiet', dest='log_level',
                       action=LogLevelAction,
                       const=LogLevelAction.DEFAULT_INCREMENT,
                       help="decrease the verbosity level")

    return group


class LogLevelAction(argparse.Action):
    """Custom argument parser's :class:`Action` for handling
    log level / verbosity flags.
    """
    DEFAULT_LEVEL = logging.WARNING

    DEFAULT_INCREMENT = 10
    DEFAULT_MINIMUM = logging.NOTSET
    DEFAULT_MAXIMUM = logging.CRITICAL

    def __init__(self, *args, **kwargs):
        self.min = kwargs.pop('min', self.DEFAULT_MINIMUM)
        self.max = kwargs.pop('max', self.DEFAULT_MAXIMUM)
        kwargs.setdefault('const', self.DEFAULT_INCREMENT)
        super(LogLevelAction, self).__init__(*args, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        current = getattr(namespace, self.dest, self.DEFAULT_LEVEL)
        new = max(self.min, min(self.max, current + self.const))
        setattr(namespace, self.dest, new)


# Miscellaneous options

def add_misc_group(parser):
    """Include the argument group with miscellaneous options.
    :param parser: :class:`argparse.ArgumentParser`
    :return: Resulting argument group
    """
    group = parser.add_argument_group("Miscellaneous", "Other options")

    group.add_argument('--version', action='version', version=__version__)
    group.add_argument('-h', '--help', action='help',
                       help="show this help message and exit")

    return group



"""
Module defining data types used in command line parsing.
"""
from enum import Enum


__all__ = ['GistCommand']


class GistCommand(Enum):
    """Command to carry out against the gist."""

    #: Run the gist's code. This should be the default.
    RUN = 'run'

    #: Output the path to gist's binary.
    WHICH = 'which'

    #: Print the complete source code of the gist's binary.
    PRINT = 'print'

    #: Open thr gist's HTML page (on GitHub) in the default web browser.
    OPEN = 'open'

    #: Display summary information about the gist.
    INFO = 'info'

    @property
    def long_flag(self):
        return '--' + self.value

    @property
    def short_flag(self):
        return '-' + self.value[0]

    @property
    def flags(self):
        return [self.short_flag, self.long_flag]

    @classmethod
    def for_flag(cls, flag):
        """Return :class:`GistCommand` matching given command line flag."""
        for cmd in cls:
            if flag in (cmd.short_flag, cmd.long_flag):
                return cmd



"""
Module with our custom extensions & enhancements
to the third party libraries used by the application.
"""
from datetime import datetime, timedelta
from pathlib import Path
import pickle

from hammock import Hammock
import requests

from gisht.util import ensure_path


__all__ = ['CachedHammock']


class CachedHammock(Hammock):
    """Custom version of :class:`Hammock` REST client
    that supports caching of GET requests.
    """
    def __init__(self, *args, **kwargs):
        """Constructor.

        :param cache_dir: Directory where the cached responses will be stored
        :param cache_ttl: :class:`timedelta` with TTL for a cache item
        """
        self._cache_dir = kwargs.pop('cache_dir', None)
        self._cache_ttl = kwargs.pop('cache_ttl', timedelta(days=7))
        super(CachedHammock, self).__init__(*args, **kwargs)

    def _path(self):
        """Return only the path portion of the URL."""
        result = "/" if self._append_slash else ""
        path_component = self
        while path_component._parent:
            result = "/" + path_component._name + result
            path_component = path_component._parent
        return result[1:] if len(result) > 1 else result  # no leading slash

    def _request(self, method, *args, **kwargs):
        """Make the HTTP request using :module:`requests` module."""
        use_cache = (method.upper() == 'GET' and
                     bool(self._cache_dir) and
                     kwargs.pop('cache', True))

        # try to load the response from cache, if available
        if use_cache:
            url_path = self._chain(*args)._path()
            cache_file = Path(self._cache_dir) / url_path
            if self._expired(cache_file):
                rv = self._on_cache_miss(url_path)
                if rv is not None:
                    return rv
            else:
                with cache_file.open('rb') as f:
                    cached_response = pickle.load(f)

                # register the cache hit and possibly modify cached response
                rv = self._on_cache_hit(url_path, cached_response)
                if rv is not False:
                    return cached_response if rv in (None, True) else rv

        # issue the request if we couldn't find a cached response
        try:
            response = super(CachedHammock, self) \
                ._request(method, *args, **kwargs)
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout,
                requests.exceptions.RetryError):
            # if the request has failed due to transient error,
            # by default return the cached response even if it's expired
            if use_cache and cache_file.exists():
                with cache_file.open('rb') as f:
                    cached_response = pickle.load(f)
                rv = self._on_cache_rescue(url_path, cached_response)
                if rv is not False:
                    return cached_response if rv in (None, True) else rv
            raise

        # save the obtained response to cache
        if use_cache:
            ensure_path(cache_file.parent)
            with cache_file.open('wb') as f:
                pickle.dump(response, f)

        return response

    def _expired(self, cache_file):
        """Check whether the cached response has expired."""
        if not cache_file.exists():
            return True
        last_modify_timestamp = cache_file.stat().st_mtime
        last_modify_time = datetime.fromtimestamp(last_modify_timestamp)
        return last_modify_time + self._cache_ttl <= datetime.now()

    # Caching-related to events
    # (to override in subclasses, if desired)

    def _on_cache_miss(self, path):
        """Invoked when given ``path`` does not have a cached response.

        :return: Anything other than ``None`` will be used in lieu
                 of the cached response (but it will **not** be saved
                 in the actual cache).
        """

    def _on_cache_hit(self, path, content):
        """Invoked when given request ``path`` has a non-stale ``content``
        that can be returned as a response to the request.

        :return: ``True`` or ``None`` if ``content`` shall be used as response.
                 ``False`` if cache shall not be used for this request at all.
                 Any other value will replace ``content`` as the response.
        """

    def _on_cache_rescue(self, path, content):
        """Invoked when given request ``path`` cannot be accessed due
        to network connectivity error, but there is cached ``content``
        available.

        :return: ``True`` or ``None`` if ``content`` shall be used as response.
                 ``False`` if the error shall be propagated.
                 Any other value will replace ``content`` as the response.
        """



"""
Module implementing requests to GitHub API.
"""
from collections import OrderedDict

import requests

from gisht import CACHE_DIR, flags, logger
from gisht.args.data import GistCommand
from gisht.ext import CachedHammock
from gisht.util import error


__all__ = ['get_gist_info', 'iter_gists']


def get_gist_info(gist_id):
    """Retrieve information about gist of given ID.

    :param gist_id: ID of a GitHub gist
                    (NOT the user-visible <owner>/<name> string!)

    :return: Dictionary with gist information
    :raises: :class:`requests.exception.HTTPError`
    """
    gist_id = str(gist_id)
    if '/' in gist_id:
        raise ValueError("expected gist ID, not the <owner>/<name> reference!")

    github = GitHub()
    response = github.gists.GET(gist_id)
    response.raise_for_status()
    return response.json()


def iter_gists(owner):
    """Iterate over gists owned by given user.

    :param owner: GitHub user's name
    :return: Iterable (generator) of parsed JSON objects (dictionaries)
    :raises: :class:`requests.exception.HTTPError`
    """
    if type(owner).__name__ not in ('str', 'unicode'):
        raise TypeError("expected a string")

    def generator():
        github = GitHub()
        gists_url = str(github.users(owner).gists)
        while gists_url:
            gists_response = requests.get(
                gists_url, params={'per_page': GitHub.RESPONSE_PAGE_SIZE})
            gists_response.raise_for_status()

            for gist_json in to_json(gists_response):
                yield gist_json

            gists_url = gists_response.links.get('next', {}).get('url')

    return generator()


# API client

class GitHub(CachedHammock):
    """Client for GitHub REST API."""
    API_URL = 'https://api.github.com'

    #: Size of the GitHub response page in items (e.g. gists).
    RESPONSE_PAGE_SIZE = 50

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('cache_dir', CACHE_DIR / 'github')
        super(GitHub, self).__init__(self.API_URL, *args, **kwargs)

    def _on_cache_miss(self, path):
        if flags.local:
            error("can't access GitHub path /%s in --local mode", path)

    def on_cache_hit(self, path, content):
        if flags.local is False:
            return False  # bypass the cache

    def _on_cache_rescue(self, path, content):
        if flags.command == GistCommand.INFO:
            logger.warning("could not communicate with GitHub -- "
                           "gist information may be out of date")


# Utility functions

def to_json(response):
    """Interpret given Requests' response object as JSON."""
    return response.json(object_pairs_hook=OrderedDict)



"""
Package for handling command line arguments.
"""
from gisht.args import autocomplete
from gisht.args.parser import create_argv_parser


__all__ = ['parse_argv']


def parse_argv(argv, namespace=None):
    """Parse command line arguments.

    :param argv: List of command line argument strings,
                 *including* the program name in argv[0]
    :param namespace: `argparse`\ 's :class:`Namespace` object
                      to fill with parsed arguments

    :return: Parse result from :func:`argparse.ArgumentParser.parse_args`
    """
    # treat everything after -- in the command line as arguments to be passed
    # to the gist executable itself (rather than be parsed by us)
    # TODO(xion): implement this as custom argparse action with nargs=REMAINDER
    gist_args = []
    if '--' in argv:
        double_dash_pos = len(argv) - 1 - argv[::-1].index('--')  # last occur.
        gist_args = argv[double_dash_pos + 1:]
        argv = argv[:double_dash_pos]

    parser = create_argv_parser()

    # trigger autocompletion, unless gist args have been provided after --
    # (otherwise they may be considered as part of GIST argument)
    # TODO(xion): make ArgumentParser handle GIST_ARGS and remove this hack
    if not gist_args:
        autocomplete.trigger(parser)

    # TODO(xion): support reading default parameter values from ~/.gishtrc
    result = parser.parse_args(argv[1:], namespace)
    result.gist_args = gist_args
    return result



"""
Module with logic for generating autocomplete suggestions
to command line arguments.
"""
import argcomplete
import requests

from gisht import BIN_DIR
from gisht.github import iter_gists


__all__ = [
    'trigger',
    'gist_completer',
]


def trigger(parser):
    """Trigger autocomplete for given :class:`ArgumentParser`,
    if the application was invoked for this purpose.
    """
    argcomplete.autocomplete(parser)


def gist_completer(prefix, parsed_args, **kwargs):
    """Autocompleter for the GIST command line argument.

    It tries to complete the gist identifier (<owner>/<name>) by listing
    all the gists for the typed <owner>, possibly filtered
    through a prefix of <name>.

    :return: Iterable of possible completions
    """
    results = set()

    # start with the locally available gists, possibly including entries
    # for GitHub users whose gists we have cached (if autocomplete prefix
    # does not include a slash)
    for entry in BIN_DIR.rglob('*'):
        entry = '/'.join(entry.relative_to(BIN_DIR).parts)
        if not entry.startswith(prefix):
            continue
        if '/' not in entry:
            entry = entry + '/'
        results.add(entry)
    # TODO(xion): the above is somewhat redundant in typical case, when GitHub
    # is available and user typed the owner part fully; elide it in this case

    # if username was given in full, query GitHub for that user's gist
    # and include them in the suggestions
    local = getattr(parsed_args, 'local', False)
    if '/' in prefix and not local:
        owner, name_prefix = prefix.split('/')
        try:
            for gist_json in iter_gists(owner):  # TODO(xion): cache this LOL
                for filename in gist_json['files'].keys():
                    if filename.startswith(name_prefix):
                        results.add(owner + '/' + filename)
                        break  # only add one (first) file per gist
        except requests.exceptions.HTTPError:
            pass  # limit to local completions if GitHub is unreachable

    return sorted(results)
